"""
A simulation runs trades against Curve pools, possibly a variety of
informed and noisy trades, although currently only "optimal" arbitrages
are supported.

The primary use-case is to determine optimal amplitude (A) and
fee parameters given historical price and volume feeds.
"""
from curvesim.pipelines.arbitrage import DEFAULT_PARAMS, volume_limited_arbitrage
from curvesim.pool_data import get as get_pool_data


def autosim(pool=None, chain="mainnet", pool_data=None, **kwargs):
    """
    The autosim() function simulates existing Curve pools with a range of
    parameters (e.g., the amplitude parameter, A, and/or the exchange fee).

    The function fetches pool properties (e.g., current pool size) and 2
    months of price/volume data, runs multiple simulations in parallel, and
    saves results plots to the "results" directory.

    Curve pools from any chain supported by the Convex Community Subgraphs
    can be simulated directly by inputting the pool's address or its LP
    token symbol.

    Parameters
    ----------
    pool: str, optional
        This string identifies the pool by address or LP token symbol.

        .. note::
            Either `pool` or `pool_data` must be provided.

        .. warning::
            An LP token symbol need not be unique.  In particular, factory
            pools are deployed permissionlessly and no checks are done to
            ensure unique LP token symbol.  Currently the first pool
            retrieved from the subgraph is used, which can be effectively
            random if token symbols clash.

    chain: str, default='mainnet'
        Identifier for blockchain or layer2.  Supported values are:
            "mainnet", "arbitrum", "optimism", "fantom", "avalanche"
            "matic", "xdai"

    pool_data: PoolData, optional
        Pool data necessary to instantiate a pool object.

        .. note::
            Either `pool` or `pool_data` must be provided.

    A: int or iterable of int, optional
        Amplification coefficient.  This controls the curvature of the
        stableswap bonding curve.  Increased values makes the curve
        flatter in a greater neighborhood of equal balances.

        Defaults to ``[int(2 ** (a / 2)) for a in range(12, 28)]``.

        For basepool, use **A_base**.

    D: int, optional
        Total pool liquidity given in 18 decimal precision.
        Defaults to on-chain data.

        For basepool, use **D_base**.

    tokens: int, optional
        Total LP token supply.
        Defaults to on-chain data.

        For basepool, use **tokens_base**.

    fee: int or iterable of int, optional
        Fees taken for both liquidity providers and the DAO.

        Units are in fixed-point so that 10**10 is 100%,
        e.g. 4 * 10**6 is 4 bps and 2 * 10**8 is 2%.

        Defaults to ``list(range(1000000, 5000000, 1000000))``.

        For basepool, use **fee_base**.

    fee_mul : int
        Fee multiplier for dynamic fee pools.

        For basepool, use **fee_mul_base**.

    admin_fee : int, default=0 * 10**9
        Fees taken for the DAO.  For factory pools, it is half of the total
        fees, as was typical for previous non-factory pools.

        Units are fixed-point percentage of `fee`, e.g. 5 * 10**9 is
        50% of the total fees.

    test: bool, default=False
        Overrides variable_params to use four test values:

        .. code-block::

            {"A": [100, 1000], "fee": [3000000, 4000000]}

    days: int, default=60
        Number of days to fetch data for.

    src: str, default='coingecko'
        Valid values for data source are 'coingecko', 'nomics', or 'local'

    data_dir: str, default='data'
        Relative path to saved data folder.

    vol_mult : float or numpy.ndarray, default computed from data
        Value(s) multiplied by market volume to specify volume limits
        (overrides vol_mode).

        Can be a scalar or vector with values for each pairwise coin
        combination.

    vol_mode : int, default=1
        Modes for limiting trade volume.
        1: limits trade volumes proportionally to market volume for each pair
        2: limits trade volumes equally across pairs
        3: mode 2 for trades with meta-pool asset, mode 1 for basepool-only trades

    ncpu : int, default=os.cpu_count()
        Number of cores to use.

    Returns
    -------
    dict
        Dictionary of results, each value being a pandas.Series.
    """
    assert any([pool, pool_data]), "Must input 'pool' or 'pool_data'"

    pool_data = pool_data or get_pool_data(pool, chain)
    p_var, p_fixed, kwargs = _parse_arguments(**kwargs)

    results = volume_limited_arbitrage(
        pool_data, variable_params=p_var, fixed_params=p_fixed, **kwargs
    )

    return results


def _parse_arguments(**kwargs):
    pool_args = ["A", "D", "x", "p", "fee", "fee_mul", "tokens", "admin_fee"]
    basepool_args = [arg + "_base" for arg in pool_args[:-1]]

    variable_params = {}
    fixed_params = {}
    rest_of_params = {}
    defaults = DEFAULT_PARAMS.copy()

    for key, val in kwargs.items():
        if key in defaults:
            del defaults[key]

        if key in pool_args:
            if isinstance(val, int):
                fixed_params.update({key: val})

            elif all(isinstance(v, int) for v in val):
                variable_params.update({key: val})

            else:
                raise TypeError(f"Argument {key} must be an int or iterable of ints")

        elif key in basepool_args:
            if isinstance(val, int):
                fixed_params.setdefault("basepool", {})
                fixed_params["basepool"].update({key[:-5]: val})

            elif all(isinstance(v, int) for v in val):
                variable_params.setdefault("basepool", {})
                variable_params["basepool"].update({key[:-5]: val})

            else:
                raise TypeError(f"Argument {key} must be an int or iterable of ints")
        else:
            rest_of_params[key] = val

    variable_params = variable_params or defaults

    return variable_params, fixed_params, rest_of_params



from math import prod

from gmpy2 import mpz


def get_xp(x, p):
    return [x * p // 10**18 for x, p in zip(x, p)]


def get_virtual_price(xp, A, tokens):
    return get_D(xp, A) * 10**18 // tokens


def get_D(xp, A):
    """
    D invariant calculation in non-overflowing integer operations
    iteratively

    A * sum(x_i) * n**n + D = A * D * n**n + D**(n+1) / (n**n * prod(x_i))

    Converging solution:
    D[j+1] = (A * n**n * sum(x_i) - D[j]**(n+1) / (n**n prod(x_i))) / (A * n**n - 1)
    """
    Dprev = 0
    S = sum(xp)
    D = S
    n = len(xp)
    Ann = A * n
    D = mpz(D)
    Ann = mpz(Ann)
    while abs(D - Dprev) > 1:
        D_P = D
        for x in xp:
            D_P = D_P * D // (n * x)
        Dprev = D
        D = (Ann * S + D_P * n) * D // ((Ann - 1) * D + (n + 1) * D_P)

    D = int(D)
    return D


def get_D_mem(rates, balances, A, xp=None):
    xp = xp or get_xp(balances, rates)
    return get_D(xp, A)


def exchange(i, j, dx, x, p, A, fee, admin_fee, xp=None, fee_mul=None):
    xp = xp or get_xp(x, p)
    x_i = xp[i] + dx * p[i] // 10**18
    y = get_y(i, j, x_i, xp, A)
    dy = xp[j] - y - 1

    if fee_mul is None:
        fee = dy * fee // 10**10
    else:
        fee = (
            dy
            * dynamic_fee((xp[i] + x_i) // 2, (xp[j] + y) // 2, fee, fee_mul)
            // 10**10
        )

    admin_fee = fee * admin_fee // 10**10

    # Convert all to real units
    rate = p[j]
    dy = (dy - fee) * 10**18 // rate
    fee = fee * 10**18 // rate
    admin_fee = admin_fee * 10**18 // rate
    assert dy >= 0

    x = x[:]
    x[i] += dx
    x[j] -= dy + admin_fee

    return x, dy, fee, admin_fee


def exchange_underlying(
    i,
    j,
    dx,
    x,
    x_base,
    rates,
    rates_base,
    A,
    A_base,
    max_coin,
    tokens_base,
    fee,
    fee_base,
    admin_fee,
    base_xp=None,
):
    # Use base_i or base_j if they are >= 0
    base_i = i - max_coin
    base_j = j - max_coin
    meta_i = max_coin
    meta_j = max_coin
    if base_i < 0:
        meta_i = i
    if base_j < 0:
        meta_j = j

    if base_i < 0 or base_j < 0:  # if i or j not in basepool
        xp = [x * p // 10**18 for x, p in zip(x, rates)]

        if base_i < 0:
            x_i = xp[i] + dx * rates[i] // 10**18
        else:
            # i is from BasePool
            # At first, get the amount of pool tokens
            n_base = len(x_base)
            base_inputs = [0] * n_base
            base_inputs[base_i] = dx
            # Deposit and measure delta
            x_base, tokens_base, dx, _ = add_liquidity(
                base_inputs,
                A_base,
                x_base,
                rates_base,
                tokens_base,
                fee_base,
                admin_fee,
                xp=base_xp,
            )

            # Need to convert pool token to "virtual" units using rates
            x_i = dx * rates[max_coin] // 10**18
            # Adding number of pool tokens
            x_i += xp[max_coin]

        y = get_y(meta_i, meta_j, x_i, xp, A)

        # Either a real coin or token
        dy = xp[meta_j] - y - 1
        dy_fee = dy * fee // 10**10

        # Convert all to real units
        # Works for both pool coins and real coins
        dy = (dy - dy_fee) * 10**18 // rates[meta_j]

        dy_admin_fee = dy_fee * admin_fee // 10**10
        dy_admin_fee = dy_admin_fee * 10**18 // rates[meta_j]

        dy_fee = dy_fee * 10**18 // rates[meta_j]

        # Change balances exactly in same way as we change actual ERC20 coin amounts
        x = x[:]
        x[meta_i] += dx
        # When rounding errors happen, we undercharge admin fee in favor of LP
        x[meta_j] -= dy + dy_admin_fee

        # Withdraw from the base pool if needed
        if base_j >= 0:
            x_base, tokens_base, dy, dy_fee, admin_fee = remove_liquidity_one_coin(
                dy,
                base_j,
                A_base,
                x_base,
                rates_base,
                tokens_base,
                fee_base,
                admin_fee,
                xp=base_xp,
            )
    else:
        # If both are from the base pool
        x_base, dy, dy_fee, dy_admin_fee = exchange(
            base_i,
            base_j,
            dx,
            x_base,
            rates_base,
            A_base,
            fee_base,
            admin_fee,
            xp=base_xp,
        )

    return x, x_base, tokens_base, dy, dy_fee, dy_admin_fee


def get_y(i, j, xp_i, xp, A):
    """
    Calculate x[j] if one makes x[i] = x

    Done by solving quadratic equation iteratively.
    x_1**2 + x1 * (sum' - (A * n - 1) * D / (A * n)) = D**(n + 1)/(n**(n + 1) * prod' * A)
    x_1**2 + b*x_1 = c

    x_1 = (x_1**2 + c) / (2*x_1 + b)
    """
    n = len(xp)
    xx = xp[:]
    D = get_D(xx, A)
    D = mpz(D)
    xx[i] = xp_i  # x is quantity of underlying asset brought to 1e18 precision
    xx = [xx[k] for k in range(n) if k != j]
    Ann = A * n
    c = D
    for y in xx:
        c = c * D // (y * n)
    c = c * D // (n * Ann)
    b = sum(xx) + D // Ann - D
    y_prev = 0
    y = D
    while abs(y - y_prev) > 1:
        y_prev = y
        y = (y**2 + c) // (2 * y + b)
    y = int(y)
    return y  # result is in units for D


def dynamic_fee(xpi, xpj, fee, fee_mul):
    xps2 = xpi + xpj
    xps2 *= xps2  # Doing just ** 2 can overflow apparently
    return (fee_mul * fee) // ((fee_mul - 10**10) * 4 * xpi * xpj // xps2 + 10**10)


def dydx_metapool(
    i,
    j,
    x,
    x_base,
    rates,
    rates_base,
    A,
    A_base,
    max_coin,
    tokens_base,
    fee=None,
    fee_base=None,
    base_xp=None,
    dx=10**12,
):
    """
    Returns price, dy[j]/dx[i], given some dx[i]

    For metapools, the indices are assumed to include base pool
    underlyer indices.

    --------------------------------
    -- Metapool pricing formula ----
    --------------------------------
    z: primary coin balance
    w: basepool virtual balance
    x_i: basepool coin balances

    dz/dx_i = dz/dw  * dw/dx_i = dz/dw * dD/dx_i = dz/dw * D'
    where D refers to the basepool

    D' = -1 * ( A * n ** (n+1) * prod(x_k) + D ** (n+1) / x_i)
            / ( n ** n * prod(x_k) - A * n ** (n+1) * prod(x_k) - (n + 1) * D ** n
    """
    base_i = i - max_coin
    base_j = j - max_coin
    base_xp = base_xp or [mpz(x) * p // 10**18 for x, p in zip(x_base, rates_base)]

    # both in and out tokens are in basepool
    if base_i >= 0 and base_j >= 0:
        _dydx = dydx(base_i, base_j, base_xp, A_base, fee=fee_base)
        return float(_dydx)

    xp = [mpz(x) * p // 10**18 for x, p in zip(x, rates)]

    x_prod = prod(base_xp)
    n = len(base_xp)
    D = mpz(get_D(base_xp, A_base))
    D_pow = D ** (n + 1)
    A_pow = A_base * n ** (n + 1)

    if base_i < 0:  # i is primary
        xj = base_xp[base_j]
        D_prime = (
            -1
            * (A_pow * x_prod + D_pow / xj)
            / (n**n * x_prod - A_pow * x_prod - (n + 1) * D**n)
        )
        D_prime = float(D_prime)

        dwdz = dydx(0, max_coin, xp, A, fee=fee)

        _dydx = dwdz / D_prime

        if fee_base:
            fee = fee_base - fee_base * xj // sum(base_xp) + 5 * 10**5
        else:
            fee = 0
        _dydx *= 1 - fee / 10**10

    else:  # i is from basepool
        base_inputs = [0] * n
        base_inputs[base_i] = dx * 10**18 // rates_base[base_i]

        dw = calc_token_amount(
            base_inputs, A_base, x_base, rates_base, tokens_base, fee=fee_base
        )
        if fee_base:
            dw = dw[0]
        # Convert lp token amount to virtual units
        dw = dw * rates[max_coin] // 10**18
        x = xp[max_coin] + dw

        meta_i = max_coin
        meta_j = j
        y = get_y(meta_i, meta_j, x, xp, A)

        dy = xp[meta_j] - y - 1
        if fee:
            dy_fee = dy * fee // 10**10
        else:
            dy_fee = 0
        dy -= dy_fee

        _dydx = dy / dx

    return float(_dydx)


def dydx(i, j, xp, A, fee=None, fee_mul=None, **kwargs):
    """
    Treats indices as applying to the "top-level" pool if a metapool.
    Basically this is the "regular" pricing calc with no special metapool handling.
    """
    xi = xp[i]
    xj = xp[j]
    n = len(xp)
    D = get_D(xp, A)
    D_pow = mpz(D) ** (n + 1)
    x_prod = prod(xp)
    A_pow = A * n ** (n + 1)
    dydx = (xj * (xi * A_pow * x_prod + D_pow)) / (xi * (xj * A_pow * x_prod + D_pow))

    if fee:
        if fee_mul is None:
            fee_factor = fee / 10**10
        else:
            dx = 10**12
            fee_factor = (
                dynamic_fee(xi + dx // 2, xj - int(dydx * dx) // 2, fee, fee_mul)
                / 10**10
            )
    else:
        fee_factor = 0

    dydx *= 1 - fee_factor

    return float(dydx)


def dydx_metapool_rai(*args, **kwargs):
    i = args[0]
    j = args[1]
    rates = args[4]
    rates_base = args[5]
    max_coin = args[8]

    dydx = dydx_metapool(*args, **kwargs)

    if i >= max_coin and j < max_coin:
        base_i = i - max_coin
        dydx = dydx * rates_base[base_i] / rates[j]

    if j >= max_coin and i < max_coin:
        base_j = j - max_coin
        dydx = dydx * rates[i] / rates_base[base_j]

    return dydx


def dydx_rai(*args, p=None, **kwargs):
    i = args[0]
    j = args[1]
    _dydx = dydx(*args, **kwargs)
    return _dydx * p[i] / p[j]


def add_liquidity(amounts, A, x, rates, tokens, fee, admin_fee, xp=None):
    mint_amount, fees = calc_token_amount(amounts, A, x, rates, tokens, fee=fee, xp=xp)
    tokens += mint_amount

    admin_fees = [f * admin_fee // 10**10 for f in fees]

    new_balances = [bal + amt - fee for bal, amt, fee in zip(x, amounts, admin_fees)]

    return new_balances, tokens, mint_amount, admin_fees


def remove_liquidity_one_coin(
    token_amount, i, A, x, p, tokens, fee, admin_fee, xp=None
):
    xp = xp or get_xp(x, p)
    dy, dy_fee = calc_withdraw_one_coin(token_amount, i, A, xp, p, tokens, fee=fee)

    admin_fee = dy_fee * admin_fee // 10**10

    x = x[:]
    x[i] -= dy + admin_fee
    tokens -= token_amount

    return x, tokens, dy, dy_fee, admin_fee


def calc_token_amount(amounts, A, x, rates, tokens, fee=False, xp=None):
    """
    Fee logic is based on add_liquidity, which makes this more accurate than
    the `calc_token_amount` in the actual contract, which neglects fees.

    By default, it's assumed you want the contract behavior.
    """
    n = len(x)
    old_balances = x
    D0 = get_D_mem(rates, old_balances, A, xp=xp)

    new_balances = x[:]
    for i in range(n):
        new_balances[i] += amounts[i]
    D1 = get_D_mem(rates, new_balances, A)

    mint_balances = new_balances[:]

    if fee:
        _fee = fee * n // (4 * (n - 1))

        fees = [0] * n
        for i in range(n):
            ideal_balance = D1 * old_balances[i] // D0
            difference = abs(ideal_balance - new_balances[i])
            fees[i] = _fee * difference // 10**10
            mint_balances[i] -= fees[i]

    D2 = get_D_mem(rates, mint_balances, A)

    mint_amount = tokens * (D2 - D0) // D0
    if fee:
        return mint_amount, fees
    else:
        return mint_amount


def calc_withdraw_one_coin(token_amount, i, A, xp, p, tokens, fee=False):
    A = A
    n = len(xp)
    D0 = get_D(xp, A)
    D1 = D0 - token_amount * D0 // tokens

    new_y = get_y_D(A, i, xp, D1)
    dy_before_fee = (xp[i] - new_y) * 10**18 // p[i]

    xp_reduced = xp
    if fee:
        n_coins = n
        _fee = fee * n_coins // (4 * (n_coins - 1))

        for j in range(n_coins):
            dx_expected = 0
            if j == i:
                dx_expected = xp[j] * D1 // D0 - new_y
            else:
                dx_expected = xp[j] - xp[j] * D1 // D0
            xp_reduced[j] -= _fee * dx_expected // 10**10

    dy = xp[i] - get_y_D(A, i, xp_reduced, D1)
    dy = (dy - 1) * 10**18 // p[i]
    if fee:
        dy_fee = dy_before_fee - dy
        return dy, dy_fee
    else:
        return dy


def get_y_D(A, i, xp, D):
    """
    Calculate x[j] if one makes x[i] = x

    Done by solving quadratic equation iteratively.
    x_1**2 + x1 * (sum' - (A*n**n - 1) * D / (A * n**n)) = D ** (n+1)/(n ** (2 * n) * prod' * A)
    x_1**2 + b*x_1 = c

    x_1 = (x_1**2 + c) / (2*x_1 + b)
    """
    n = len(xp)
    D = mpz(D)
    xx = [xp[k] for k in range(n) if k != i]
    S = sum(xx)
    Ann = A * n
    c = D
    for y in xx:
        c = c * D // (y * n)
    c = c * D // (n * Ann)
    b = S + D // Ann
    y_prev = 0
    y = D
    while abs(y - y_prev) > 1:
        y_prev = y
        y = (y**2 + c) // (2 * y + b - D)
    y = int(y)
    return y  # result is in units for D



from multiprocessing import Pool as cpu_pool
from types import MethodType


def run_pipeline(param_sampler, price_sampler, strategy, ncpu=4):
    """
    Core function for running pipelines.

    Typically called within a function specifying the pipeline components
    (see, e.g., :func:`.volume_limited_arbitrage()`)

    Parameters
    ----------
    param_sampler : iterator
        An iterator that returns pool parameters (see :mod:`.param_samplers`).

    price_sampler : iterator
        An iterator that returns (minimally) a time-series of prices
        (see :mod:`.price_samplers`).

    strategy: callable
        A function dictating what happens at each timestep.

    ncpu : int, default=4
        Number of cores to use.

    Returns
    -------
    results : tuple
        Contains the metrics produced by the strategy.

    """
    if ncpu > 1:
        price_sampler_data = list(price_sampler)
        args = [(pool, params, price_sampler_data) for pool, params in param_sampler]

        with cpu_pool(ncpu) as clust:
            results = zip(*clust.starmap(strategy, args))

    else:
        results = []
        for pool, params in param_sampler:
            metrics = strategy(pool, params, price_sampler.restart())
            results.append(metrics)
        results = tuple(zip(*results))

    return results


class SimInterface:
    """
    A template class for creating simulation interfaces for pools.
    See pool.stablewap.interfaces.StableSwapSimInterface

    This component is likely to change.
    """

    def __init__(self):
        self.pool = None
        self.coin_indices = None
        self.pool_state = None

    @staticmethod
    def _get_pool_state(pool):  # pylint: disable=method-hidden
        raise NotImplementedError

    @staticmethod
    def _init_coin_indices(metadata):  # pylint: disable=method-hidden
        raise NotImplementedError

    def price(self, coin_in, coin_out, use_fee=True):
        raise NotImplementedError

    def test_trade(self, coin_in, coin_out, dx, state=None):
        raise NotImplementedError

    def _set_pool_interface(self, pool, pool_function_dict):
        """
        Binds the pool and functions used in simulation to the interface.

        Parameters
        ----------
        pool :
            A pool object.

        pool_function_dict : dict
            A dict with interface method names as keys, and functions as values.

            Note: Currently, _get_pool_state, and _init_coin_indices are required.

        """
        self.pool = pool
        pool_function_dict = pool_function_dict.copy()

        # Set Required Functions
        self._get_pool_state = pool_function_dict.pop("_get_pool_state")
        self._init_coin_indices = pool_function_dict.pop("_init_coin_indices")

        # Set Additional Functions
        for interface_fn, pool_fn in pool_function_dict.items():
            bound_method = MethodType(pool_fn, self)
            setattr(self, interface_fn, bound_method)

        self.coin_indices = self._init_coin_indices(self.pool.metadata)
        self.set_pool_state()

    def get_pool_state(self):
        """
        Gets pool state using the provided _get_pool_state method
        """
        return self._get_pool_state(self.pool)

    def set_pool_state(self):
        """
        Records the current pool state in the interface's pool_state atttribute.
        """
        self.pool_state = self.get_pool_state()

    def get_coin_indices(self, *coins):
        """
        Gets the pool indices for the input coin names.
        Uses the coin_indices set by _init_coin_indices.
        """
        coin_indices = self.coin_indices
        return [self._get_coin_index(coin_indices, c) for c in coins]

    @staticmethod
    def _get_coin_index(coin_indices, coin_id):
        """
        Gets the index for a single coin based on its name.
        Uses the coin_indices set by _init_coin_indices.
        """
        if isinstance(coin_id, str):
            coin_id = coin_indices[coin_id]
        return coin_id



"""
General utility for http requests.
"""
from aiohttp import ClientSession
from tenacity import retry, stop_after_attempt, wait_random

stop_rule = stop_after_attempt(5)
wait_rule = wait_random(min=0.5, max=1.5)


class HTTP:
    @staticmethod
    @retry(stop=stop_rule, wait=wait_rule)
    async def get(url, params=None):

        kwargs = {"url": url, "headers": {"Accept-Encoding": "gzip"}}

        if params is not None:
            kwargs.update({"params": params})

        async with ClientSession() as session:
            async with session.get(**kwargs) as resp:
                r = await resp.json()

        return r

    @staticmethod
    @retry(stop=stop_rule, wait=wait_rule)
    async def post(url, json=None):
        kwargs = {"url": url, "headers": {"Accept-Encoding": "gzip"}}

        if json is not None:
            kwargs.update({"json": json})

        async with ClientSession() as session:
            async with session.post(**kwargs) as resp:
                r = await resp.json()

        return r



"""
Plotters to visualize simulation results.

This code is likely to be changed soon.
"""
__all__ = ["saveplots"]

import os

import matplotlib.pyplot as plt

from .legacy import plotsims, plotsimsfee


def saveplots(folder_name, A_list, fee_list, results):
    """
    Save charts generated from `autosim`.

    Parameters
    ----------
    folder_name: str
        name of folder in `results` to save charts to;
        will create if needed
    A_list: list of int
        list of Amplification (A) paramters
    fee_list: list of int
        list of fee parameters
    results: dict
        results dict from `autosim`
    """
    base_filepath = os.path.join("results", folder_name)
    os.makedirs(base_filepath, exist_ok=True)

    if len(fee_list) > 1:
        filepath = os.path.join(base_filepath, "summary")
        plotsimsfee(
            A_list=A_list,
            fee_list=fee_list,
            ar=results["ar"],
            bal=results["bal"],
            depth=results["depth"],
            volume=results["volume"],
            show=False,
            saveas=filepath,
        )

    for curr_fee in fee_list:
        filename = "fee_" + str(round(curr_fee) / 10**8)[2:].ljust(2, "0")
        filepath = os.path.join(base_filepath, filename)

        plotsims(
            A_list=A_list,
            ar=results["ar"].loc[(slice(None), curr_fee), :],
            bal=results["bal"].loc[(slice(None), curr_fee), :],
            pool_value=results["pool_value"].loc[(slice(None), curr_fee), :],
            depth=results["depth"].loc[(slice(None), curr_fee), :],
            volume=results["volume"].loc[(slice(None), curr_fee), :],
            log_returns=results["log_returns"].loc[(slice(None), curr_fee), :],
            err=results["err"].loc[(slice(None), curr_fee), :],
            show=False,
            saveas=filepath,
        )

        plt.close("all")



"""
Module housing the required implementation for any pool entity
in the Curvesim Framework.
"""


class Pool:
    """
    The `Pool` base class has the explicitly required properties for any pool-like
    object used in Curvesim.

    Currently the base attributes are not informative for pools
    constructed manually rather than from chain data.
    """

    @property
    def name(self):
        """Descriptive name for this pool"""
        if hasattr(self, "metadata"):
            return self.metadata["name"]
        return "Custom pool"

    @property
    def address(self):
        """Address on chain"""
        if hasattr(self, "metadata"):
            return self.metadata["address"]
        return "0x" + "0" * 40

    @property
    def chain(self):
        """Chain for this pool"""
        if hasattr(self, "metadata"):
            return self.metadata["chain"]
        return "No chain"

    @property
    def pool_type(self):
        """Type of this pool"""
        if hasattr(self, "metadata"):
            return self.metadata["pool_type"]
        return "Custom"

    @property
    def symbol(self):
        """LP token symbol"""
        if hasattr(self, "metadata"):
            return self.metadata["symbol"]
        return "Custom"

    def __repr__(self):
        return f"<{self.__class__.__name__} address={self.address} chain={self.chain}>"



# pylint: disable=redefined-outer-name
import os

import boa
import pytest

_base_dir = os.path.dirname(__file__)
_curve_dir = os.path.join(_base_dir, "curve")
FAKE_ADDRESS = "0xCAFECAFECAFECAFECAFECAFECAFECAFECAFECAFE"


@pytest.fixture(scope="session")
def mainnet_3pool_state():
    """Snapshot of Mainnet 3Pool values"""
    p = [10**18, 10**30, 10**30]
    balances = [
        295949605740077243186725223,
        284320067518878,
        288200854907854,
    ]
    virtual_balances = [b * p // 10**18 for b, p in zip(balances, p)]

    return {
        "N_COINS": 3,
        "A": 2000,
        "p": p,
        "balances": balances,
        "virtual_balances": virtual_balances,
        "lp_tokens": 849743149250065202008212976,
        "virtual_price": 1022038799187029697,
    }


@pytest.fixture(scope="session")
def _vyper_3pool(mainnet_3pool_state):
    """Initialize vyper fixture using mainnet values."""
    lp_total_supply = mainnet_3pool_state["lp_tokens"]
    mock_filepath = os.path.join(_base_dir, "lp_token_mock.vy")
    lp_token = boa.load(mock_filepath, lp_total_supply)

    pool_filepath = os.path.join(_curve_dir, "basepool.vy")
    owner = FAKE_ADDRESS
    coins = [FAKE_ADDRESS] * 3
    A = mainnet_3pool_state["A"]
    fee = 4 * 10**6
    admin_fee = 5 * 10**9
    pool = boa.load(pool_filepath, owner, coins, lp_token, A, fee, admin_fee)

    balances = mainnet_3pool_state["balances"]
    pool.eval(f"self.balances={balances}")

    return pool


@pytest.fixture(scope="session")
def _vyper_metapool(_vyper_3pool):
    """Initialize vyper fixture using mainnet values."""
    metapool_filepath = os.path.join(_curve_dir, "metapool.vy")
    name = "SIM-3Pool"
    symbol = "SIM3CRV-f"
    coin = FAKE_ADDRESS
    rate_multiplier = 10**34  # 2 decimals
    basepool = _vyper_3pool.address
    basepool_token = _vyper_3pool.token()
    A = 1000
    fee = 4 * 10**6
    # Admin fee is hard-coded as 50% for factory pools
    # admin_fee = 5 * 10**9
    metapool = boa.load(
        metapool_filepath,
        name,
        symbol,
        coin,
        basepool,
        basepool_token,
        rate_multiplier,
        A,
        fee,
    )

    balances = [762951074, 12971664836474542835562756]
    metapool.eval(f"self.balances={balances}")
    total_supply = 20312687702458911532611097
    metapool.eval(f"self.totalSupply={total_supply}")

    return metapool


@pytest.fixture(scope="function")
def vyper_3pool(_vyper_3pool):
    with boa.env.anchor():
        yield _vyper_3pool


@pytest.fixture(scope="function")
def vyper_metapool(_vyper_metapool):
    with boa.env.anchor():
        yield _vyper_metapool



"""
Tools for fetching pool state and metadata.
Currently supports stableswap pools, meta-pools, and rebasing (RAI) metapools.
"""

__all__ = ["PoolData", "from_address", "from_symbol", "get"]

from numpy import array

from ..network.subgraph import redemption_prices_sync as _redemption_prices
from ..network.subgraph import volume_sync as _volume
from ..pool.stableswap import CurveMetaPool, CurvePool, CurveRaiPool
from .queries import from_address, from_symbol


def get(address_or_symbol, chain="mainnet"):
    """
    Pulls pool state and metadata from daily snapshot.

    Parameters
    ----------
    address_or_symbol : str
        Pool address prefixed with “0x” or LP token symbol.

    chain : str
        Chain/layer2 identifier, e.g. “mainnet”, “arbitrum”, “optimism".

    Returns
    -------
    PoolData

    """
    if address_or_symbol.startswith("0x"):
        from_x = from_address
    else:
        from_x = from_symbol

    params = from_x(address_or_symbol, chain)

    pool_data = PoolData(params)

    return pool_data


class PoolData:
    """
    Container with methods to return pool state, metadata, and pools employing them.
    """

    def __init__(self, metadata_dict, cache_data=False, days=60):
        """
        Parameters
        ----------
        metadata_dict : dict
            Pool metadata in the format returned by network.subgraph.pool_snapshot.

        cache_data : bool, optional
            If True, fetches and caches historical volume and redemption price.

        days : int, default=60
            Number of days to pull data for if caching.

        """
        self.dict = metadata_dict
        self._volume = None
        self._redemption_prices = None

        if cache_data:
            self.set_cache(days=days)

    def set_cache(self, days=60):
        """
        Fetches and caches historical volume and redemption price data.

        Parameters
        ----------
        days : int, default=60
            number of days to pull data for
        """
        self.volume(days=days, store=True)
        self.redemption_prices(store=True)

    def clear_cache(self):
        """
        Clears any cached data.
        """
        self._volume = None
        self._redemption_prices = None

    def pool(self, balanced=True, balanced_base=True):
        """
        Constructs a pool object based on the stored data.

        Parameters
        ----------
        balanced : bool, default=True
            If True, balances the pool value across assets.

        balanced_base : bool, default=True
            If True and pool is metapool, balances the basepool value across assets.

        Returns
        -------
        Pool
        """

        def bal(kwargs, balanced):
            reserves = kwargs.pop("reserves")
            if not balanced:
                kwargs.update({"D": reserves})
            return kwargs

        kwargs = bal(self.dict["init_kwargs"].copy(), balanced)

        if self.dict["basepool"]:
            bp_kwargs = self.dict["basepool"]["init_kwargs"].copy()
            bp_kwargs = bal(bp_kwargs, balanced_base)
            kwargs.update({"basepool": CurvePool(**bp_kwargs)})

            r = self.redemption_prices()
            if r is None:
                pool = CurveMetaPool(**kwargs)
            else:
                pool = CurveRaiPool(r, **kwargs)

        else:
            pool = CurvePool(**kwargs)

        pool.metadata = self.dict

        return pool

    def coins(self):
        """
        Returns coin addresses for the pool's holdings.

        For pools that are not on Ethereum mainnet, the address
        for the corresponding mainnet token is returned.

        For lending tokens (e.g., aTokens or cTokens), the
        address for the underlying token is returned.

        Returns
        -------
        list of strings
            coin addresses

        """
        if not self.dict["basepool"]:
            c = self.dict["coins"]["addresses"]
        else:
            c = (
                self.dict["coins"]["addresses"][:-1]
                + self.dict["basepool"]["coins"]["addresses"]
            )
        return c

    def coin_names(self):
        """
        Returns coin names for the pool's holdings.

        For pools that are not on Ethereum mainnet, the name
        of the corresponding mainnet token is returned.

        For lending tokens (e.g., aTokens or cTokens), the
        name of the underlying token is returned.

        Returns
        -------
        list of strings
            coin names

        """
        if not self.dict["basepool"]:
            c = self.dict["coins"]["names"]
        else:
            c = (
                self.dict["coins"]["names"][:-1]
                + self.dict["basepool"]["coins"]["names"]
            )
        return c

    def volume(self, days=60, store=False, get_cache=True):
        """
        Fetches the pool's historical volume over the specified number of days.

        Parameters
        ----------
        days : int, default=60
            Number of days to pull data for.

        store : bool, default=False
            If true, caches the fetched data.

        get_cache : bool, default=True
            If true, returns cached data when available.

        Returns
        -------
        numpy.ndarray
            Total volume summed across the specified number of days.

        """
        if get_cache and self._volume is not None:
            print("Getting cached historical volume...")
            return self._volume

        print("Fetching historical volume...")
        addrs = self.dict["address"]
        chain = self.dict["chain"]

        if self.dict["basepool"]:
            addrs = [addrs, self.dict["basepool"]["address"]]
            vol = _volume(addrs, chain, days=days)
            summed_vol = array([sum(v) for v in vol])

        else:
            vol = _volume(addrs, chain, days=days)
            summed_vol = array(sum(vol))

        if store:
            self._volume = summed_vol

        return summed_vol

    def n(self):
        """
        Returns the number of token-types (e.g., DAI, USDC, USDT) in a pool.

        Returns
        -------
        int or list of ints
            Number of token-types.

            For metapools, a list [n_metapool, n_basepool] is returned.

            N_metapool includes the basepool LP token.

        """
        if not self.dict["basepool"]:
            n = self.dict["init_kwargs"]["n"]
        else:
            n = [
                self.dict["init_kwargs"]["n"],
                self.dict["basepool"]["init_kwargs"]["n"],
            ]

        return n

    def type(self):
        """
        Returns the pool type.

        Returns
        -------
        str

        """
        if self.dict["basepool"]:
            return CurveMetaPool
        else:
            return CurvePool

    def redemption_prices(self, days=60, store=False, get_cache=True):
        """
        Fetches the pool's redemption price over the specified number of days.

        Note: only returns data for RAI3CRV pool. Otherwise, returns None.

        Parameters
        ----------
        days : int, default=60
            Number of days to pull data for.

        store : bool, default=False
            If True, caches the fetched data.

        get_cache : bool, default=True
            If True, returns cached data when available.

        Returns
        -------
        pandas.DataFrame
            Timestamped redemption prices across the specified number of days.

        """
        if get_cache and self._redemption_prices is not None:
            print("Getting cached redemption prices...")
            return self._redemption_prices

        address = self.dict["address"]
        chain = self.dict["chain"]

        r = _redemption_prices(address, chain, days=days)

        if store:
            self._redemption_prices = r

        return r



"""
Pool subpackage for creating Curve pool objects.

This subpackage can be used standalone from the simulation functionality
for interactive exploration of Curve pool behavior. The two primary ways to create
pools are `get` and `make`, which allow you to create pools from on-chain state or p
assed-in parameters, respectively.

Pool calculations are thoroughly unit-tested against the corresponding
smart contract code.  We aim for the exact results as in the vyper
integer arithmetic. The pool interfaces largely adhere to the smart contracts
but in a few cases allow an extra option, such as enabling/disabling fees.
"""

__all__ = [
    "Pool",
    "get",
    "make",
    "CurvePool",
    "CurveMetaPool",
    "CurveRaiPool",
]

from curvesim.pool_data import get as _get_pool_data

from .base import Pool
from .stableswap import CurveMetaPool, CurvePool, CurveRaiPool


def make(
    A,
    D,
    n,
    basepool=None,
    p=None,
    tokens=None,
    fee=4 * 10**6,
    fee_mul=None,
    admin_fee=0 * 10**9,
):
    """
    Factory function for creating pools from "raw" parameters, i.e. no
    data pulled from chain.

    Parameters
    ----------
    A: int
        Amplification coefficient.  This controls the curvature of the
        stableswap bonding curve.  Increased values makes the curve
        flatter in a greater neighborhood of equal balances.

        Defaults to [int(2 ** (a / 2)) for a in range(12, 28)].

    D: int
        Total pool liquidity given in 18 decimal precision.

    n: int
        The number of token-types in the pool (e.g., DAI, USDC, USDT = 3)

    basepool: dict, optional
        a dict cointaining the arguments for instantiating a basepool

    p: list of int, optional
        precisions for each coin

    tokens: int, optional
        Total LP token supply.

        .. note::
            The number of tokens does not influence "normal" pool computations;
            but, for metapools, the number of basepool tokens is critically
            important to trade calculations.

    fee: int
        Fees taken for both liquidity providers and the DAO.

        Units are in fixed-point so that 10**10 is 100%,
        e.g. 4 * 10**6 is 4 bps and 2 * 10**8 is 2%.

    fee_mul : int
        fee multiplier for dynamic fee pools

    admin_fee : int, default=0 * 10**9
        Fees taken for the DAO.  For factory pools, it is half of the total fees,
        as was typical for previous non-factory pools.

        Units are fixed-point percentage of `fee`, e.g. 5 * 10**9 is
        50% of the total fees.

    Returns
    -------
    :class:`Pool`
    """
    if basepool:
        pool = CurveMetaPool(
            A,
            D,
            n,
            basepool,
            p=p,
            tokens=tokens,
            fee=fee,
            fee_mul=fee_mul,
            admin_fee=admin_fee,
        )

    else:
        pool = CurvePool(
            A,
            D,
            n,
            p=p,
            tokens=tokens,
            fee=fee,
            fee_mul=fee_mul,
            admin_fee=admin_fee,
        )

    return pool


def get(address_or_symbol, chain="mainnet", balanced=False, balanced_base=False):
    """
    Parameters
    ----------
    address_or_symbol: str
        pool address prefixed with "0x" or LP token symbol

        .. warning::
            An LP token symbol need not be unique.  In particular, factory pools
            are deployed permissionlessly and no checks are done to ensure unique
            LP token symbol.  Currently the first pool retrieved from the subgraph
            is used, which can be effectively random if token symbols clash.

    chain: str, default="mainnet"
        chain/layer2 identifier, e.g. "mainnet", "arbitrum", "optimism"

    balanced : bool, default=False
            If True, balances the pool value across assets.

    balanced_base : bool, default=False
            If True and pool is metapool, balances the basepool value across assets.

    Returns
    -------
    :class:`Pool`

    Examples
    --------
    >>> import curvesim
    >>> pool = curvesim.pool.get("0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7", "mainnet")
    """
    p = _get_pool_data(address_or_symbol, chain=chain)
    return p.pool(balanced=balanced, balanced_base=balanced_base)



class CurvesimException(Exception):
    """Base exception class"""


class SubgraphError(CurvesimException):
    """Raised for errors in subgraph query"""


class SubgraphResultError(SubgraphError):
    """Raised when subgraph results aren't as expected"""


class MissingEnvVarError(CurvesimException, RuntimeError):
    """Environment variable is missing."""



from datetime import datetime, timedelta

from curvesim.network import coingecko as _coingecko
from curvesim.network import nomics as _nomics


def nomics(coins, days=60, data_dir="data"):
    t_end = datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
    t_start = t_end - timedelta(days=days)

    print("Fetching Nomics price data...")
    print("Timerange: %s to %s" % (str(t_start), str(t_end)))

    _nomics.update(coins, None, t_start, t_end, data_dir=data_dir)
    prices, volumes, pzero = _nomics.pool_prices(coins)

    return prices, volumes, pzero


def coingecko(coins, days=60):
    print("Fetching CoinGecko price data...")
    prices, volumes = _coingecko.pool_prices(coins, "usd", days)
    pzero = 0

    return prices, volumes, pzero


def local(coins, data_dir="data"):
    print("Using local data...")
    prices, volumes, pzero = _nomics.local_pool_prices(coins, data_dir=data_dir)

    return prices, volumes, pzero



"""
Plotters to visualize simulation results.

This code is likely to be changed soon.
"""


import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
import numpy as np
from matplotlib.lines import Line2D


def plotsims(
    *,
    A_list,
    ar,
    bal,
    pool_value,
    depth,
    volume,
    log_returns,
    err,
    show=True,
    saveas=False,
):
    """
    Plots output of Asims when only 1 fee is used
    """
    _plotsims_page_1(
        A_list=A_list,
        ar=ar,
        bal=bal,
        depth=depth,
        volume=volume,
        err=err,
    )

    if saveas:
        plt.savefig(saveas + "_1.png")

    _plotsims_page_2(
        A_list=A_list,
        bal=bal,
        pool_value=pool_value,
        depth=depth,
        volume=volume,
        log_returns=log_returns,
        err=err,
    )

    if saveas:
        plt.savefig(saveas + "_2.png")

    if show:
        plt.show()


def _plotsims_page_1(
    *,
    A_list,
    ar,
    bal,
    depth,
    volume,
    err,
):
    colors = plt.cm.viridis(np.linspace(0, 1, len(A_list)))

    # Summary stats
    _, axs = plt.subplots(2, 3, constrained_layout=True, figsize=(8, 5))

    axs[0, 0].plot(ar.unstack(level=1) * 100, "k", zorder=1)
    axs[0, 0].scatter(A_list, ar * 100, c=colors, zorder=2)
    axs[0, 0].yaxis.set_major_formatter(mtick.PercentFormatter())
    axs[0, 0].set_xlabel("Amplitude (A)")
    axs[0, 0].set_ylabel("Annualized Returns")

    axs[0, 1].plot(A_list, np.median(depth, axis=1), "k", zorder=1, label="Med")
    axs[0, 1].plot(A_list, np.min(depth, axis=1), "k--", zorder=1, label="Min")
    axs[0, 1].scatter(A_list, np.median(depth, axis=1), c=colors, zorder=2)
    axs[0, 1].scatter(A_list, np.min(depth, axis=1), c=colors, zorder=2)
    axs[0, 1].set_xlabel("Amplitude (A)")
    axs[0, 1].set_ylabel("Liquidity Density")
    axs[0, 1].legend(loc="lower right")

    axs[0, 2].plot(A_list, bal.median(axis=1), "k", zorder=1, label="Med")
    axs[0, 2].plot(A_list, bal.min(axis=1), "k--", zorder=1, label="Min")
    axs[0, 2].scatter(A_list, bal.median(axis=1), c=colors, zorder=2)
    axs[0, 2].scatter(A_list, bal.min(axis=1), c=colors, zorder=2)
    axs[0, 2].set_ylim([0, 1])
    axs[0, 2].set_xlabel("Amplitude (A)")
    axs[0, 2].set_ylabel("Pool Balance")
    axs[0, 2].legend(loc="lower right")

    axs[1, 0].plot(A_list, volume.sum(axis=1) / 60, "k", zorder=1)
    axs[1, 0].scatter(A_list, volume.sum(axis=1) / 60, c=colors, zorder=2)
    axs[1, 0].set_xlabel("Amplitude (A)")
    axs[1, 0].set_ylabel("Daily Volume")

    axs[1, 1].plot(A_list, err.median(axis=1), "k", zorder=1)
    axs[1, 1].scatter(A_list, err.median(axis=1), c=colors, zorder=2)
    axs[1, 1].set_xlabel("Amplitude (A)")
    axs[1, 1].set_ylabel("Median Price Error")

    # Legend
    handles = []
    for c in colors:
        line = Line2D(
            [0],
            [0],
            marker="o",
            color="w",
            markerfacecolor=c,
            markersize=10,
        )
        handles.append(line)

    axs[1, 2].legend(handles, A_list, title="Amplitude", ncol=2)
    axs[1, 2].axis("off")


def _plotsims_page_2(
    *,
    A_list,
    bal,
    pool_value,
    depth,
    volume,
    log_returns,
    err,
):
    colors = plt.cm.viridis(np.linspace(0, 1, len(A_list)))

    # Time-series Data
    _, axs = plt.subplots(3, 2, constrained_layout=True, figsize=(8, 5))

    # Pool value
    for i, color in enumerate(colors):
        axs[0, 0].plot(pool_value.iloc[i], color=color)

    axs[0, 0].set_ylabel("Pool Value")
    plt.setp(axs[0, 0].xaxis.get_majorticklabels(), rotation=40, ha="right")
    axs[0, 0].yaxis.get_major_formatter().set_useOffset(False)

    # Balance
    for i, color in enumerate(colors):
        axs[0, 1].plot(bal.iloc[i], color=color)

    axs[0, 1].set_ylabel("Pool Balance")
    plt.setp(axs[0, 1].xaxis.get_majorticklabels(), rotation=40, ha="right")

    # Volume
    for i, color in enumerate(colors):
        axs[1, 0].plot(volume.T.resample("1D").sum().T.iloc[i], color=color)

    axs[1, 0].set_ylabel("Daily Volume")
    plt.setp(axs[1, 0].xaxis.get_majorticklabels(), rotation=40, ha="right")

    # Depth
    for i, color in enumerate(colors):
        axs[1, 1].plot(depth.iloc[i], color=color)

    axs[1, 1].set_ylabel("Liquidity Density")
    plt.setp(axs[1, 1].xaxis.get_majorticklabels(), rotation=40, ha="right")

    # Distribution of log returns
    axs[2, 0].hist(log_returns.T, 30, histtype="step", color=colors)
    axs[2, 0].set_xlabel("Log Returns")
    axs[2, 0].set_ylabel("Frequency")

    # Price error
    axs[2, 1].hist(err.T, 30, histtype="step", color=colors)
    axs[2, 1].set_xlabel("Price Error")
    axs[2, 1].set_ylabel("Frequency")


def plotsimsfee(*, A_list, fee_list, ar, bal, depth, volume, show=True, saveas=False):
    """
    Plots 2D summary output of Asims when multiple fees are used

    """
    fig, axs = plt.subplots(2, 3, constrained_layout=True, figsize=(11, 5.5))
    fee_list_pct = np.array(fee_list) / 10**8

    def plot_subfigure(ax, im, title, xlabel, ylabel):
        ax.set_title(title)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_xticks(np.arange(len(fee_list)))
        ax.set_yticks(np.arange(len(A_list)))
        ax.set_xticklabels(fee_list_pct)
        ax.set_yticklabels(A_list)
        plt.setp(ax.xaxis.get_majorticklabels(), rotation=90)
        cbar = fig.colorbar(im, ax=ax)
        return cbar

    # Annualized Returns
    ax = axs[0, 0]
    im = ax.imshow(ar.unstack("fee") * 100, cmap="plasma")
    plot_subfigure(
        ax, im, title="Annualized Returns (%)", xlabel="Fee (%)", ylabel="Amplitude (A)"
    )

    # Volume
    ax = axs[1, 0]
    im = ax.imshow(volume.sum(axis=1).unstack("fee"), cmap="plasma")
    plot_subfigure(ax, im, title="Volume", xlabel="Fee (%)", ylabel="Amplitude (A)")

    # Median Depth
    ax = axs[0, 1]
    im = ax.imshow(depth.median(axis=1).unstack("fee"), cmap="plasma")
    cbar = plot_subfigure(
        ax,
        im,
        title="Median Liquidity Density",
        xlabel="Fee (%)",
        ylabel="Amplitude (A)",
    )
    cbar.ax.yaxis.set_major_formatter(mtick.PercentFormatter())

    # Minimum Depth
    ax = axs[1, 1]
    im = ax.imshow(depth.min(axis=1).unstack("fee"), cmap="plasma")
    cbar = plot_subfigure(
        ax,
        im,
        title="Minimum Liquidity Density",
        xlabel="Fee (%)",
        ylabel="Amplitude (A)",
    )
    cbar.ax.yaxis.set_major_formatter(mtick.PercentFormatter())

    # Median Balance
    ax = axs[0, 2]
    im = ax.imshow(bal.median(axis=1).unstack("fee"), cmap="plasma")
    plot_subfigure(
        ax, im, title="Median Balance", xlabel="Fee (%)", ylabel="Amplitude (A)"
    )

    # Minimum Balance
    ax = axs[1, 2]
    im = ax.imshow(bal.min(axis=1).unstack("fee"), cmap="plasma")
    plot_subfigure(
        ax, im, title="Minimum Balance", xlabel="Fee (%)", ylabel="Amplitude (A)"
    )

    if saveas:
        plt.savefig(saveas + ".png")

    if show:
        plt.show()



from collections import namedtuple
from itertools import combinations

from gmpy2 import mpz
from numpy import isnan

from curvesim.pipelines.templates import SimInterface

from . import functions as pool_functions
from .metapool import CurveMetaPool
from .pool import CurvePool
from .raipool import CurveRaiPool


class StableSwapSimInterface(SimInterface):
    def __init__(self, pool):
        super().__init__()

        pool_function_dict = stableswap_interface_fns[type(pool)]
        self._set_pool_interface(pool, pool_function_dict)

        self.pricing_fns = stableswap_pricing_fns[type(pool)]
        self.next_timestamp = self.pool.next_timestamp

        all_idx = range(pool.n_total)
        base_idx = list(range(pool.n))
        self.max_coin = getattr(pool, "max_coin", None)

        if self.max_coin:
            base_idx[self.max_coin] = "bp_token"

        self.index_combos = list(combinations(all_idx, 2))
        self.base_index_combos = list(combinations(base_idx, 2))

    def get_liquidity_density(self, coin_in, coin_out, factor=10**8):
        # Fix: won't work for trades between meta-pool and basepool
        i, j = self.get_coin_indices(coin_in, coin_out)
        state = self.get_pool_state()

        x = getattr(state, "x_base", state.x)
        p = getattr(state, "p_base", state.p)

        if i == "bp_token":
            i = self.max_coin
            x = state.x
            p = state.rates

        if j == "bp_token":
            j = self.max_coin
            x = state.x
            p = state.rates

        xp = pool_functions.get_xp(x, p)

        price_pre = self.price(coin_in, coin_out)
        output = self.test_trade(coin_in, coin_out, xp[i] // factor, state=state)
        price_post = output[0]
        LD1 = price_pre / ((price_pre - price_post) * factor)

        price_pre = self.price(coin_out, coin_in)
        output = self.test_trade(coin_out, coin_in, xp[j] // factor, state=state)
        price_post = output[0]
        LD2 = price_pre / ((price_pre - price_post) * factor)

        return (LD1 + LD2) / 2


# Namedtuples for pool states
PoolState = namedtuple(
    "PoolState", ["x", "p", "A", "fee", "fee_mul", "tokens", "admin_fee"]
)


MetaPoolState = namedtuple(
    "MetaPoolState",
    [
        "x",
        "x_base",
        "p",
        "p_base",
        "A",
        "A_base",
        "fee",
        "fee_base",
        "fee_mul",
        "fee_mul_base",
        "tokens",
        "tokens_base",
        "admin_fee",
        "rates",
    ],
)


# Functions for stableswap.CurvePool
def _get_pool_state(p):
    return PoolState(p.x[:], p.p[:], p.A, p.fee, p.fee_mul, p.tokens, p.admin_fee)


def _precisions(self):
    state = self.get_pool_state()
    return state.p


def _init_pool_coin_indices(metadata):
    coin_names = metadata["coins"]["names"]
    coin_indices = range(len(coin_names))
    return dict(zip(coin_names, coin_indices))


def _price(self, coin_in, coin_out, use_fee=True):
    i, j = self.get_coin_indices(coin_in, coin_out)
    assert i != j
    return self.pool.dydx(i, j, use_fee=use_fee)


def _trade(self, coin_in, coin_out, size):
    i, j = self.get_coin_indices(coin_in, coin_out)
    assert i != j

    output = self.pool.exchange(i, j, size)
    self.set_pool_state()

    return output


def _test_trade(self, coin_in, coin_out, dx, state=None):
    i, j = self.get_coin_indices(coin_in, coin_out)
    assert i != j

    state = state or self.get_pool_state()
    exchange_args = (state.x, state.p, state.A, state.fee, state.admin_fee)

    output = pool_functions.exchange(i, j, dx, *exchange_args, fee_mul=state.fee_mul)

    xp_post = [x * p // 10**18 for x, p in zip(output[0], state.p)]

    dydx = pool_functions.dydx(
        i, j, xp_post, state.A, fee=state.fee, fee_mul=state.fee_mul
    )

    return (dydx,) + output


def _make_error_fns(self):
    # Note: for performance, does not support string coin-names

    state = self.get_pool_state()
    args = [state.x, state.p, state.A, state.fee, state.admin_fee]
    xp = pool_functions.get_xp(state.x, state.p)

    def get_trade_bounds(i, j):
        xp_j = int(xp[j] * 0.01)
        high = pool_functions.get_y(j, i, xp_j, xp, state.A) - xp[i]

        return (0, high)

    def post_trade_price_error(dx, i, j, price_target):
        dx = int(dx) * 10**18 // state.p[i]

        if dx > 0:
            output = pool_functions.exchange(
                i, j, dx, xp=xp, *args, fee_mul=state.fee_mul
            )

            xp_post = pool_functions.get_xp(output[0], state.p)
        else:
            xp_post = xp

        dydx = pool_functions.dydx(
            i, j, xp_post, state.A, fee=state.fee, fee_mul=state.fee_mul
        )

        return dydx - price_target

    def post_trade_price_error_multi(dxs, price_targets, coins):
        _args = args[:]
        _xp = xp

        # Do trades
        for k, pair in enumerate(coins):
            i, j = pair

            if isnan(dxs[k]):
                dx = 0
            else:
                dx = int(dxs[k]) * 10**18 // state.p[i]

            if dx > 0:
                output = pool_functions.exchange(
                    i, j, dx, *_args, fee_mul=state.fee_mul, xp=_xp
                )

                _args[0] = output[0]  # update x
                _xp = pool_functions.get_xp(_args[0], state.p)

        # Record price errors
        errors = []
        for k, pair in enumerate(coins):
            dydx = pool_functions.dydx(
                *pair, _xp, state.A, fee=state.fee, fee_mul=state.fee_mul
            )
            errors.append(dydx - price_targets[k])

        return errors

    return get_trade_bounds, post_trade_price_error, post_trade_price_error_multi


# Functions for stableswap.CurveMetaPool
def _get_metapool_state(pool):
    state_pairs = zip(_get_pool_state(pool), _get_pool_state(pool.basepool))

    args = [arg for pair in state_pairs for arg in pair]
    args[-1] = pool.rates()
    args = args

    return MetaPoolState(*args)


def _precisions_metapool(self):
    state = self.get_pool_state()
    return state.p[0 : self.max_coin] + state.p_base


def _init_metapool_coin_indices(metadata):
    meta_coin_names = metadata["coins"]["names"][:-1]
    base_coin_names = metadata["basepool"]["coins"]["names"]

    coin_names = meta_coin_names + base_coin_names
    coin_indices = range(len(coin_names))
    coin_dict = dict(zip(coin_names, coin_indices))

    bp_token_name = metadata["coins"]["names"][-1]
    bp_token_dict = dict.fromkeys([bp_token_name, "bp_token"], "bp_token")
    coin_dict.update(bp_token_dict)

    return coin_dict


def _metapool_price(self, coin_in, coin_out, use_fee=True):
    i, j = self.get_coin_indices(coin_in, coin_out)

    # Basepool LP token not used
    if i != "bp_token" and j != "bp_token":
        assert i != j
        return self.pool.dydx(i, j, use_fee=use_fee)

    # Basepool LP token used
    else:
        if i == "bp_token":
            i = self.max_coin

        if j == "bp_token":
            j = self.max_coin

        assert i != j
        xp = self.pool._xp()
        return self.pool._dydx(i, j, xp=xp, use_fee=use_fee)


def _metapool_trade(self, i, j, size):
    pool = self.pool
    trade_fn = pool.exchange_underlying

    if isinstance(i, str):
        i = self.coin_indices[i]
        if i == "bp_token":
            i = self.max_coin
            trade_fn = pool.exchange

    if isinstance(j, str):
        j = self.coin_indices[j]
        if j == "bp_token":
            j = self.max_coin
            trade_fn = pool.exchange

    assert i != j
    output = trade_fn(i, j, size)
    self.set_pool_state()

    return output


def _metapool_test_trade(self, coin_in, coin_out, dx, state=None):
    i, j = self.get_coin_indices(coin_in, coin_out)
    state = state or self.get_pool_state()
    max_coin = self.max_coin
    get_dydx, _get_dydx = self.pricing_fns

    # Basepool LP token not used in trade
    if i != "bp_token" and j != "bp_token":
        assert i != j
        return _test_trade_underlying(state, get_dydx, i, j, dx, max_coin)

    # Basepool LP token used in trade
    else:
        if i == "bp_token":
            i = max_coin

        if j == "bp_token":
            j = max_coin

        assert i != j
        return _test_trade_meta(state, _get_dydx, i, j, dx)


def _test_trade_underlying(state, pricing_fn, i, j, dx, max_coin):
    args = [
        state.x,
        state.x_base,
        state.rates,
        state.p_base,
        state.A,
        state.A_base,
        max_coin,
        state.tokens_base,
        state.fee,
        state.fee_base,
    ]

    output = pool_functions.exchange_underlying(
        i, j, dx, *args, admin_fee=state.admin_fee
    )

    # Update x values
    args[0:2] = output[0:2]

    # Update rates and tokens if needed
    if i < max_coin or j < max_coin:
        tokens_base = output[2]
        xp_base = pool_functions.get_xp(output[1], state.p_base)
        vp_base = pool_functions.get_virtual_price(xp_base, state.A_base, tokens_base)

        rates = state.rates[:]
        rates[max_coin] = vp_base

        args[2] = rates
        args[7] = tokens_base

    dydx = pricing_fn(i, j, *args)

    return (dydx,) + output


def _test_trade_meta(state, pricing_fn, i, j, dx):
    xp = pool_functions.get_xp(state.x, state.rates)
    exchange_args = (
        state.x,
        state.p,
        state.A,
        state.fee,
        state.admin_fee,
        xp,
        state.fee_mul,
    )

    output = pool_functions.exchange(
        i,
        j,
        dx,
        *exchange_args,
    )

    xp_post = pool_functions.get_xp(output[0], state.rates)

    dydx = pricing_fn(
        i, j, xp_post, state.A, p=state.p, fee=state.fee, fee_mul=state.fee_mul
    )

    return (dydx,) + output


def _make_metapool_error_fns(self):  # noqa: C901
    # Note: for performance, does not support string coin-names

    state = self.get_pool_state()
    get_dydx = self.pricing_fns[0]
    max_coin = self.max_coin
    args = [
        state.x,
        state.x_base,
        state.rates,
        state.p_base,
        state.A,
        state.A_base,
        max_coin,
        state.tokens_base,
        state.fee,
        state.fee_base,
    ]

    p_all = state.p[0:max_coin] + state.p_base
    xp_meta = pool_functions.get_xp(state.x, state.rates)
    xp_base = pool_functions.get_xp(state.x_base, state.p_base)

    def get_trade_bounds(i, j):
        base_i = i - max_coin
        base_j = j - max_coin
        meta_i = max_coin
        meta_j = max_coin
        if base_i < 0:
            meta_i = i
        if base_j < 0:
            meta_j = j

        if base_i < 0 or base_j < 0:
            xp_j = int(xp_meta[meta_j] * 0.01)
            high = pool_functions.get_y(meta_j, meta_i, xp_j, xp_meta, state.A)
            high -= xp_meta[meta_i]
            # high = high * 10**18 // state.rates[meta_i]
        else:
            xp_j = int(xp_base[base_j] * 0.01)
            high = pool_functions.get_y(base_j, base_i, xp_j, xp_base, state.A_base)
            high -= xp_base[base_i]
            # high = high * 10**18 // state.p_base[base_i]

        return (0, high)

    def post_trade_price_error(dx, i, j, price_target):
        _args = args[:]

        dx = int(dx) * 10**18 // p_all[i]

        if dx > 0:
            output = pool_functions.exchange_underlying(
                i, j, dx, *_args, admin_fee=state.admin_fee, base_xp=xp_base
            )

            # Update x, x_base, rates and tokens
            tokens_base = output[2]
            xp_base_post = [x * p // 10**18 for x, p in zip(output[1], state.p_base)]
            vp_base = pool_functions.get_virtual_price(
                xp_base_post, state.A_base, tokens_base
            )

            rates = state.rates[:]
            rates[max_coin] = vp_base

            _args[0:3] = output[0:2] + (rates,)
            _args[7] = tokens_base

        else:
            xp_base_post = xp_base

        dydx = get_dydx(i, j, *_args, base_xp=xp_base_post)

        return dydx - price_target

    def post_trade_price_error_multi(dxs, price_targets, coins):
        _args = args[:]
        _xp_base = xp_base

        # Do trades
        for k, pair in enumerate(coins):
            i, j = pair

            if isnan(dxs[k]):
                dx = 0
            else:
                dx = int(dxs[k]) * 10**18 // p_all[i]

            if dx > 0:
                output = pool_functions.exchange_underlying(
                    i, j, dx, *_args, admin_fee=state.admin_fee, base_xp=_xp_base
                )

                # Update x, x_base, rates and tokens
                tokens_base = output[2]
                _xp_base = [x * p // 10**18 for x, p in zip(output[1], state.p_base)]
                vp_base = pool_functions.get_virtual_price(
                    _xp_base, state.A_base, tokens_base
                )

                rates = state.rates[:]
                rates[max_coin] = vp_base

                _args[0:3] = output[0:2] + (rates,)
                _args[7] = tokens_base

        # Record price errors
        errors = []
        _xp_base = [mpz(x) for x in _xp_base]
        for k, pair in enumerate(coins):

            dydx = get_dydx(*pair, *_args, base_xp=_xp_base)
            errors.append(dydx - price_targets[k])

        return errors

    return get_trade_bounds, post_trade_price_error, post_trade_price_error_multi


# Functions to set for each pool types
interface_functions = (
    "price",
    "trade",
    "test_trade",
    "make_error_fns",
    "precisions",
    "_get_pool_state",  # required
    "_init_coin_indices",  # required
)

stableswap_pool_fns = (
    _price,
    _trade,
    _test_trade,
    _make_error_fns,
    _precisions,
    _get_pool_state,
    _init_pool_coin_indices,
)

stableswap_metapool_fns = (
    _metapool_price,
    _metapool_trade,
    _metapool_test_trade,
    _make_metapool_error_fns,
    _precisions_metapool,
    _get_metapool_state,
    _init_metapool_coin_indices,
)

stableswap_pool_fns = dict(zip(interface_functions, stableswap_pool_fns))
stableswap_metapool_fns = dict(zip(interface_functions, stableswap_metapool_fns))
stableswap_interface_fns = {
    CurvePool: stableswap_pool_fns,
    CurveMetaPool: stableswap_metapool_fns,
    CurveRaiPool: stableswap_metapool_fns,
}

stableswap_pricing_fns = {
    CurvePool: (pool_functions.dydx, pool_functions.dydx),
    CurveMetaPool: (pool_functions.dydx_metapool, pool_functions.dydx),
    CurveRaiPool: (pool_functions.dydx_metapool_rai, pool_functions.dydx_rai),
}



import os

from dotenv import load_dotenv

from curvesim.exceptions import MissingEnvVarError

load_dotenv()


# Dummy value for optional default arg
# so that any value, including `None`,
# can be set as a default.
_NOT_VALUE = object()


def get_env_var(var_name, default=_NOT_VALUE):
    """
    Retrieve environment variable.

    Parameters
    ----------
    var_name: str
        Name of the environment variable.
    default: object
        Value to return if env var is missing.

    Returns
    -------
    str
        Value of the environment variable.

    Raise
    -----
    curvesim.exception.MissingEnvVarError
        Raised if default is not set and env var is missing.
    """
    var_value = os.getenv(var_name)
    if var_value is None:
        if default is _NOT_VALUE:
            raise MissingEnvVarError(f"Could not get env var: '{var_name}'")
        else:
            return default

    return var_value



from .metapool import CurveMetaPool


class CurveRaiPool(CurveMetaPool):
    """
    Rebasing stableswap metapool implementation in Python. Used for RAI3CRV pool.

    """

    def __init__(self, redemption_prices, *args, p=None, n=None, **kwargs):
        """
        Parameters
        ----------
        redemption_prices : pandas.DataFrame
            timestamped redemption prices
            (see :meth:`.PoolData.redemption_prices()`)
        A : int
            Amplification coefficient; this is :math:`A n^{n-1}` in the whitepaper.
        D : int or list of int
            coin balances or virtual total balance
        n: int
            number of coins
        p: list of int
            precision and rate adjustments
        tokens : int
            LP token supply
        fee : int, optional
            fee with 10**10 precision (default = .004%)
        fee_mul :
            fee multiplier for dynamic fee pools
        admin_fee : int, optional
            percentage of `fee` with 10**10 precision (default = 50%)
        """
        self.redemption_prices = redemption_prices

        p = p or [10**18] * n
        p[0] = int(redemption_prices.price[0])

        super().__init__(p=p, n=n, *args, **kwargs)

    def next_timestamp(self, timestamp, *args, **kwargs):
        """
        Updates the redemption price based on the input timestamp

        Parameters
        ----------
        timestamp : datetime.datetime
            the time to sample from

        """

        r = self.redemption_prices.price.asof(timestamp)
        self.p[0] = int(r)

    def dydx(self, i, j, use_fee=False):
        _dydx = super().dydx(i, j, use_fee=use_fee)

        if i >= self.max_coin and j < self.max_coin:
            base_i = i - self.max_coin
            _dydx = _dydx * self.basepool.p[base_i] / self.p[j]

        return _dydx

    def _dydx(self, i, j, xp, use_fee=False):
        dydx = super()._dydx(i, j, xp, use_fee=use_fee)
        return dydx * self.p[i] / self.p[j]
